import sys
import os
from argparse import RawTextHelpFormatter

import aetros.const
from aetros.commands.PredictCommand import PredictCommand
from aetros.commands.StartCommand import StartCommand

commands_dict = {
    'start': StartCommand,
    'predict': PredictCommand
}
command_summaries = [
    ['start', 'Starts a training of a network in current working directory'],
    ['stop', 'Stops a training'],
    ['predict', 'Runs a prediction locally']
]

def create_main_parser():
    import argparse
    parser = argparse.ArgumentParser(formatter_class=RawTextHelpFormatter, add_help=False, prog=const.__prog__)

    pkg_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    parser.version = 'aetros %s from %s (python %s)' % (
        const.__version__, pkg_dir, sys.version[:3])

    parser.add_argument('command', nargs='?')
    parser.add_argument('--version', action='version', version='%s' % (const.__version__,))
    parser.add_argument("-h", "--help", action='store_true', help="show this help message and exit")

    # # add the general options
    # gen_opts = cmdoptions.make_option_group(cmdoptions.general_group, parser)
    # parser.add_option_group(gen_opts)

    # parser.main = True  # so the help formatter knows

    # create command listing for description
    command_summaries = [
        ['start', 'Starts a training of a network'],
        ['stop', 'Stops a tranining']
    ]
    description = [''] + ['%-27s %s' % (i, j) for i, j in command_summaries]
    parser.description = 'Please don not forget to provide API_KEY as environment variable.\nPossible commands:\n' + ('\n'.join(description))

    return parser


def parseopts(args):
    if len(args) == 0:

        description = [''] + ['%-27s %s' % (i, j) for i, j in command_summaries]
        print("usage: API_KEY='key' aetros [command]\n")
        print('Please don not forget to provide API_KEY as environment variable.\n\nPossible commands:\n' + ('\n'.join(description)))

        sys.exit(1)

    cmd_name = args[0]

    # all the args without the subcommand
    cmd_args = args[1:]

    if not os.getenv('API_KEY'):
        print aetros.const.bcolors.FAIL + "API_KEY not defined." + aetros.const.bcolors.ENDC
        sys.exit(1)

    return cmd_name, cmd_args


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    cmd_name, cmd_args = parseopts(args)

    command = commands_dict[cmd_name]()

    return command.main(cmd_args)


import os
from threading import Thread, Lock

import requests

import signal
import json
import time

from requests.auth import HTTPBasicAuth

def dict_factory(cursor, row):
    d = {}
    for idx, col in enumerate(cursor.description):
        d[col[0]] = row[idx]
    return d


class EventListener:
    def __init__(self):
        self.events = {}

    def on(self, name, callback):
        if name not in self.events:
            self.events[name] = []

        self.events[name].append(callback)

    def fire(self, name):
        if name in self.events:
            for callback in self.events[name]:
                callback()


class AetrosBackend:
    def __init__(self, job_id):
        self.event_listener = EventListener()
        self.api_token = os.getenv('API_KEY')
        self.job_id = job_id
        self.queue = []
        self.queueLock = Lock()
        self.active_syncer = True

        self.host = os.getenv('API_HOST') if os.getenv('API_HOST') else 'aetros.com'

        self.thread = Thread(target=self.syncer)
        self.thread.daemon = True
        self.thread.start()
        self.in_request = False
        self.stop_requested = False

    def syncer(self):
        while (self.active_syncer):
            self.sync()
            time.sleep(1)

    def sync(self):
        self.queueLock.acquire()

        if len(self.queue) == 0:
            self.queueLock.release()
            return

        max_messages_per_request = 20

        queue_to_save = self.queue[:max_messages_per_request]
        self.queue = self.queue[max_messages_per_request:]
        self.queueLock.release()

        # print "Sync %d/%d items to server" % (len(queue_to_save), len(self.queue))
        self.in_request = True
        try:
            response = self.post('job/sync', json={'id': self.job_id, 'items': queue_to_save})
            failed = False
            if response.status_code != 200:
                failed = True
            else:
                if response.content:
                    content = json.loads(response.content)
                    if 'not_exists' in content:
                        print "Job deleted meanwhile. Stop current job."
                        os.kill(os.getpid(), signal.SIGINT)
                    elif 'stop' in content:
                        print "Job stopped through trainer."
                        self.stop_requested = True
                        os.kill(os.getpid(), signal.SIGINT)
                    else:
                        if 'result' in content and content['result'] is True:
                            if len(self.queue) > 0:
                                # we have still some messages, so send directly next sync
                                self.sync()
                        else:
                            failed = True
                else:
                    failed = True

            if failed:
                self.queueLock.acquire()
                self.queue = queue_to_save + self.queue
                self.queueLock.release()
                print "Error in sending job information: %s" % (response.content,)

        except Exception:
            self.queueLock.acquire()
            self.queue = queue_to_save + self.queue
            self.queueLock.release()

        self.in_request = False

    def stop_syncer(self):
        self.active_syncer = False

        while self.in_request:
            time.sleep(0.05)

        self.queueLock.acquire()
        queue = self.queue[:]
        self.queueLock.release()

        if len(queue) > 0:
            print "Sending last (%d) monitoring information to server ... " % (len(queue),)
            response = self.post('job/sync', json={'id': self.job_id, 'items': queue})
            if response.status_code != 200:
                print "Error in sending job information."

    def kill(self):
        self.active_syncer = False

    def get_url(self, affix):

        url = 'http://%s/api/%s' % (self.host, affix)

        if '?' in url:
            url += '&token=' + self.api_token
        else:
            url += '?token=' + self.api_token

        return url

    def write_log(self, id, message):
        item = {'id': id, 'message': message}

        self.queueLock.acquire()
        self.queue.append({
            'type': 'log',
            'time': time.time(),
            'data': item
        })
        self.queueLock.release()

    def upload_weights(self, name, file_path):
        if not os.path.isfile(file_path):
            return

        files = {}
        files[name] = open(file_path, 'r')
        return requests.post(self.get_url('job/weights?id=%s' % (self.job_id,)), files=files)

    def get_best_weight_url(self, job_id):
        response = self.get('job/weight-best', {'id': job_id})
        return response.json()

    def get(self, url, params=None, **kwargs):
        return requests.get(self.get_url(url), params=params, **kwargs)

    def post(self, url, data=None, **kwargs):
        return requests.post(self.get_url(url), data=data, **kwargs)

    def put(self, url, data=None, **kwargs):
        return requests.put(self.get_url(url), data=data, **kwargs)

    def create_job(self, name, server_id ='local', dataset_id = None, insights = False):
        response =  self.put('job', {'networkId': name, 'serverId': server_id, 'insights': insights, 'datasetId': dataset_id})
        if response.status_code != 200:
            print("Could not create training: %s" % (response.content, ))
            return None

        return response.json()

    def get_job(self):
        response = self.get('job', {'id': self.job_id})
        if response.status_code != 200:
            print("Could not find training: %s" % (response.content,))
            return None

        return response.json()

    def get_light_job(self):
        response = self.get('job', {'id': self.job_id, 'light': 1})
        if response.status_code != 200:
            print("Could not find training: %s" % (response.content,))
            return None

        return response.json()

    def job_started(self, id, pid):
        return self.post('job/started', {'id': id, 'pid': pid})

    def job_add_status(self, statusKey, statusValue):
        item = {'statusKey': statusKey, 'statusValue': json.dumps(statusValue, allow_nan=False)}

        self.queueLock.acquire()
        self.queue.append({
            'type': 'job-status',
            'time': time.time(),
            'data': item
        })
        self.queueLock.release()

    def job_set_info_key(self, key, info):
        self.queueLock.acquire()
        self.queue.append({
            'type': 'job-info',
            'time': time.time(),
            'data': {'key': key, 'info': info}
        })
        self.queueLock.release()

    def job_add_insight(self, info, images):
        self.queueLock.acquire()
        self.queue.append({
            'type': 'job-insight',
            'time': time.time(),
            'data': {'info': info, 'images': images}
        })
        self.queueLock.release()



from __future__ import division
import hashlib
import json
import random
import shutil
import urllib
import os
from pprint import pprint

import math
import numpy as np
import sys

import requests
from PIL import Image

from aetros.utils import get_option
from network import ensure_dir

from threading import Thread, Lock
from Queue import Queue

def download_image(url, path):
    if os.path.exists(path):
        return True

    try:
        r = requests.get(url, stream=True, timeout=9)
        if r.status_code == 200:
            with open(path, 'wb') as f:
                r.raw.decode_content = True
                shutil.copyfileobj(r.raw, f)
                return True
        else:
            print("Could not download image %s, response %d" % (url, r.status_code))
    except Exception as e:
        print("Could not download image %s due to %s" % (url, e.message))

    return False

class ImageDownloaderWorker(Thread):
    def __init__(self, q, trainer, dataset, max, images, controller):
        Thread.__init__(self)
        self.q = q
        self.trainer = trainer
        self.dataset = dataset
        self.max = max
        self.images = images
        self.controller = controller

    def run(self):
        try:
            while self.controller['running']:
                self.handle(self.q.get())
                self.q.task_done()
        except KeyboardInterrupt:
            self.controller['running'] = False
            return

    def handle(self, message):
        image = message[0]

        if 'id' not in image:
            return

        local_name = image['id']
        local_image_path = '%s/%s' % (self.trainer.job_model.get_dataset_downloads_dir(self.dataset), local_name)

        try:
            ensure_dir(os.path.dirname(local_image_path))
        except Exception:
            pass

        if not os.path.isfile(local_image_path):
            if download_image(image['src'], local_image_path):
                try:
                    img = Image.open(local_image_path)
                    resize = bool(get_option(self.dataset['config'], 'resize', True))
                    if resize:
                        os.remove(local_image_path)
                        size = (int(get_option(self.dataset['config'], 'resizeWidth', 512)), int(get_option(self.dataset['config'], 'resizeHeight', 512)))
                        quality = int(float(get_option(self.dataset['config'], 'resizeCompression', 0.8)) * 100)
                        img.thumbnail(size, Image.ANTIALIAS)

                        local_image_path = os.path.splitext(local_image_path)[0]+'.jpg'
                        img.save(local_image_path, 'JPEG', quality=quality, optimize=True)
                except IOError as e:
                    print("No valid image found %s" % (local_image_path,))
                    os.remove(local_image_path)
                except KeyboardInterrupt:
                    self.controller['running'] = False

        self.images[image['id']] = local_image_path
        self.trainer.set_status('LOAD IMAGE %d/%d' % (len(self.images), self.max))

def read_image(path, size, target_shape, grayscale=False, dim_ordering=None):
    if dim_ordering is None:
        import keras.backend
        dim_ordering = keras.backend.image_dim_ordering()

    try:
        image = Image.open(path)
    except:
        return None

    # do not only resize, maintain aspect ratio
    image = image.resize(size, Image.ANTIALIAS)

    if grayscale:
        image = image.convert("L")
    else:
        image = image.convert("RGB")

    image = np.asarray(image, dtype='float32')

    if len(target_shape) > 1:
        # RGB: height, width, channel -> channel, height, width
        if grayscale:
            if dim_ordering == 'th':
                image = image.reshape((1, size[0], size[0]))
            else:
                image = image.reshape((size[0], size[0], 1))
        else:
            if dim_ordering == 'th':
                image = image.transpose(2, 0, 1)
    else:
        # L: height, width => height*width
        image = image.reshape(size[0] * size[1])

    return image

# class to read all images in the ram at once
class ImageReadWorker(Thread):
    def __init__(self, q, trainer, path, grayscale, size, images, controller):
        Thread.__init__(self)
        self.q = q
        self.trainer = trainer
        self.path = path
        self.grayscale = grayscale
        self.size = size
        self.images = images
        self.controller = controller

    def run(self):
        while self.controller['running']:
            self.handle(self.q.get())
            self.q.task_done()

    def handle(self, message):
        path, validation, category_dir = message

        try:
            if os.path.isfile(path):
                image = read_image(path, self.size, self.trainer.input_shape, grayscale=self.grayscale)
            else:
                return

            if image is None:
                return

            self.images.append([image, validation, category_dir])
        except IOError as e:
            self.trainer.logger.write('Could not open %s due to %s' % (path, e.message))
            return

class InMemoryDataGenerator():
    def __init__(self, datagen, images, classes_count, batch_size):
        self.index = -1
        self.datagen = datagen
        self.images = images
        random.shuffle(self.images)
        self.lock = Lock()

        self.classes_count = classes_count
        self.batch_size = batch_size

    def __iter__(self):
        return self

    def next(self):
        batch_x = []
        batch_y = []

        for i in range(self.batch_size):

            with self.lock:
                self.index += 1
                # reset iterator if necessary
                if self.index == len(self.images):
                    self.index = 0
                    random.shuffle(self.images)

            image, class_idx = self.images[self.index]
            image = np.copy(image) # we need to copy it, otherwise we'd operate on the same object again and again

            if self.datagen is not None:
                image = self.datagen.random_transform(image)
                image = self.datagen.standardize(image)
            else:
                image /= 255.0

            batch_x.append(image)

            y = np.zeros((self.classes_count,), dtype='float32')
            y[class_idx] = 1.
            batch_y.append(y)

        return np.array(batch_x), np.array(batch_y)


def read_images_in_memory(job_config, dataset, node, trainer):
    """
    Reads all images into memory and applies augmentation if enabled
    """
    from keras.utils import np_utils
    concurrent = 6

    dataset_config = dataset['config']
    controller = {'running': True}
    config = dataset['config']
    q = Queue(concurrent)

    size = (node['width'], node['height'])

    if node['inputType'] == 'image':
        trainer.input_shape = (1, size[0], size[1])
        grayscale = True
    elif node['inputType'] == 'image_rgb':
        grayscale = False
        trainer.input_shape = (3, size[0], size[1])
    else:
        trainer.input_shape = (size[0] * size[1],)
        grayscale = True

    result = {
        'X_train': [],
        'Y_train': [],
        'X_test': [],
        'Y_test': []
    }

    images = []
    max = 0

    path = trainer.job_model.get_dataset_downloads_dir(dataset)
    if 'path' in dataset['config']:
        path = dataset['config']['path']

    classes_count = 0
    category_map = {}
    classes = []

    try:
        for i in range(concurrent):
            t = ImageReadWorker(q, trainer, path, grayscale, size, images, controller)
            t.daemon = True
            t.start()

        for validation_or_training in ['validation', 'training']:
            if os.path.isdir(path+'/'+validation_or_training):
                for category_name in os.listdir(path+'/'+validation_or_training):
                    if os.path.isdir(path+'/'+validation_or_training+'/'+category_name):

                        if category_name not in category_map:
                            category_map[category_name] = classes_count
                            if 'classes' in dataset_config and 'category_' in category_name:
                                category_idx = int(category_name.replace('category_', ''))
                                category_map[category_name] = category_idx
                                target_category = dataset_config['classes'][category_idx]
                                classes.append(target_category['title'] or 'Class %s' % (category_idx, ))
                            else:
                                classes.append(category_name)

                            classes_count += 1

                        for id in os.listdir(path+'/'+validation_or_training+'/'+category_name):
                            file_path = os.path.join(path, validation_or_training, category_name, id)
                            q.put([file_path, validation_or_training == 'validation', category_name])
                            max += 1

        q.join()
        controller['running'] = False

        train_images = []
        test_images = []

        for v in images:
            image, validation, category_dir = v
            if validation is True:
                test_images.append([image, category_map[category_dir]])
            else:
                train_images.append([image, category_map[category_dir]])

        train_datagen = None
        augmentation = bool(get_option(dataset_config, 'augmentation', False))
        if augmentation:
            train_datagen = get_image_data_augmentor_from_dataset(dataset)
        train = InMemoryDataGenerator(train_datagen, train_images, classes_count, job_config['settings']['batchSize'])

        test = InMemoryDataGenerator(None, test_images, classes_count, job_config['settings']['batchSize'])

        nb_sample = len(train_images)
        trainer.set_generator_training_nb(nb_sample)
        trainer.set_generator_validation_nb(len(test_images))

        print ("Found %d classes, %d images (%d in training [%saugmented], %d in validation). Read all images into memory from %s" %
               (classes_count, max, len(train_images), 'not ' if augmentation is False else '', len(test_images), path))
        pprint(category_map)

        if classes_count == 0:
            print("Could not find any classes. Does the directory contains images?")
            sys.exit(1)

        trainer.output_size = classes_count
        trainer.set_job_info('classes', classes)
        trainer.classes = classes

        result['X_train'] = train
        result['Y_train'] = train
        result['X_test'] = test
        result['Y_test'] = test

        return result

    except KeyboardInterrupt:
        controller['running'] = False
        sys.exit(1)

def get_image_data_augmentor_from_dataset(dataset):
    from keras.preprocessing.image import ImageDataGenerator
    dataset_config = dataset['config']

    augShearRange = float(get_option(dataset_config, 'augShearRange', 0.1))
    augZoomRange = float(get_option(dataset_config, 'augZoomRange', 0.1))
    augHorizontalFlip = bool(get_option(dataset_config, 'augHorizontalFlip', False))
    augVerticalFlip = bool(get_option(dataset_config, 'augVerticalFlip', False))
    augRotationRange = float(get_option(dataset_config, 'augRotationRange', 0.2))

    return ImageDataGenerator(
            rescale=1./255,
            rotation_range=augRotationRange,
            shear_range=augShearRange,
            zoom_range=augZoomRange,
            horizontal_flip=augHorizontalFlip,
            vertical_flip=augVerticalFlip
    )

def read_images_keras_generator(job_config, dataset, node, trainer):
    from keras.preprocessing.image import ImageDataGenerator

    size = (node['width'], node['height'])

    if node['inputType'] == 'image':
        trainer.input_shape = (1, size[0], size[1])
        grayscale = True
    elif node['inputType'] == 'image_rgb':
        grayscale = False
        trainer.input_shape = (3, size[0], size[1])
    else:
        trainer.input_shape = (size[0] * size[1],)
        grayscale = True

    dataset_config = dataset['config']
    print("Generate image iterator in folder %s " % (dataset_config['path'],))

    augmentation = bool(get_option(dataset_config, 'augmentation', False))

    if augmentation:
        train_datagen = get_image_data_augmentor_from_dataset(dataset)
    else:
        train_datagen = ImageDataGenerator()

    train_generator = train_datagen.flow_from_directory(
            directory=os.path.join(dataset_config['path'], 'training'),
            target_size=size,
            batch_size=job_config['settings']['batchSize'],
            color_mode='grayscale' if grayscale is True else 'rgb',
            class_mode='categorical')

    classes = []
    for folderName, outputNeuron in train_generator.class_indices.iteritems():
        if dataset['type'] == 'images_search' or dataset['type'] == 'images_upload':
            category_idx = int(folderName.replace('category_', ''))
            target_category = dataset_config['classes'][category_idx]
            classes.append(target_category['title'] or 'Category %s' % (category_idx, ))
        else:
            classes.append(folderName)

    trainer.set_job_info('classes', classes)
    trainer.classes = classes

    trainer.output_size = train_generator.nb_class

    # ensure_dir(dataset_config['path'] + '/preview')

    test_datagen = ImageDataGenerator(rescale=1./255)
    validation_generator = test_datagen.flow_from_directory(
            directory=os.path.join(dataset_config['path'], 'validation'),
            # save_to_dir=dataset_config['path'] + '/preview',
            target_size=size,
            batch_size=trainer.get_batch_size(),
            color_mode='grayscale' if grayscale is True else 'rgb',
            class_mode='categorical')

    trainer.set_generator_validation_nb(validation_generator.nb_sample)
    trainer.set_generator_training_nb(train_generator.nb_sample)

    print ("Found %d classes, %d images (%d in training [%saugmented], %d in validation) in %s " %
           (len(classes), validation_generator.nb_sample+train_generator.nb_sample, train_generator.nb_sample, 'not ' if augmentation is False else '', validation_generator.nb_sample, dataset_config['path']))

    if trainer.output_size == 0:
        print("Could not find any classes. Does the directory contains images?")
        sys.exit(1)

    pprint(train_generator.class_indices)
    pprint(classes)

    return {
        'X_train': train_generator,
        'Y_train': train_generator,
        'X_test': validation_generator,
        'Y_test': validation_generator,
    }

def get_images(job_config, dataset, node, trainer):
    concurrent = 15

    q = Queue(concurrent)
    config = dataset['config']

    dir = trainer.job_model.get_dataset_downloads_dir(dataset)

    ensure_dir(dir)

    classes = config['classes']

    size = (node['width'], node['height'])

    trainer.set_status('PREPARE_IMAGES')

    if node['inputType'] == 'image':
        trainer.input_shape = (1, size[0], size[1])
    elif node['inputType'] == 'image_rgb':
        trainer.input_shape = (3, size[0], size[1])
    else:
        trainer.input_shape = (size[0] * size[1],)

    max = 0
    images = {}

    dataset_path = trainer.job_model.get_dataset_downloads_dir(dataset)
    meta_information_file = dataset_path + '/meta.json'

    classes_changed = False
    config_changed = False
    had_previous = False
    classes_md5 = hashlib.md5(json.dumps(classes)).hexdigest()

    validationFactor = 0.2

    if os.path.isdir(dataset_path):
        if os.path.isfile(meta_information_file):
            with open(meta_information_file) as f:
                meta = json.load(f)
                if meta:
                    had_previous = True
                    if 'classes_md5' in meta and meta['classes_md5'] != classes_md5:
                        classes_changed = True

                    trigger_changed = ['resize', 'resizeWidth', 'resizeHeight', 'resizeCompression']
                    for i in trigger_changed:
                        if i in meta['config'] and i in config and meta['config'][i] != config[i]:
                            config_changed = True
                else:
                    config_changed = True
        else:
            config_changed = True

    need_download = classes_changed or config_changed

    if need_download:
        if had_previous:
            print "Reset dataset and re-download images to " + dir
            if classes_changed:
                print (" .. because classes changed")
            if config_changed:
                print (" .. because settings changed")
        else:
            print "Download images to " + dir

        resize = bool(get_option(config, 'resize', True))
        if resize:
            resizeSize = (int(get_option(config, 'resizeWidth', 64)), int(get_option(config, 'resizeHeight', 64)))
            print " .. with resizing to %dx%d " % resizeSize

        # we need to donwload all images
        shutil.rmtree(dataset_path)

        controller = {'running': True}
        try:
            for category in classes:
                max += len(category['images'])

            for i in range(concurrent):
                t = ImageDownloaderWorker(q, trainer, dataset, max, images, controller)
                t.daemon = True
                t.start()

            for category_idx, category in enumerate(classes):
                for image in category['images']:
                    q.put([image, category_idx])

            q.join()
            controller['running'] = False

            def move_image(image, category = 'training'):
                if image['id'] in images and os.path.isfile(images[image['id']]):
                    target_path = dataset_path + '/%s/category_%s/%s' % (category, category_idx, os.path.basename(images[image['id']]))
                    ensure_dir(os.path.dirname(target_path))
                    os.rename(images[image['id']], target_path)

            for category_idx, category in enumerate(classes):
                random.shuffle(category['images'])
                position = int(math.ceil(len(category['images']) * validationFactor))

                ensure_dir(dataset_path + '/training')
                ensure_dir(dataset_path + '/validation')

                for image in category['images'][position:]: #test data
                    if image['id'] in images and os.path.isfile(images[image['id']]):
                        move_image(image, 'training')

                for image in category['images'][:position]: #validation data
                    if image['id'] in images and os.path.isfile(images[image['id']]):
                        move_image(image, 'validation')

            with open(meta_information_file, 'w') as f:
                meta = {
                    'loaded_at': classes_md5,
                    'classes_md5': classes_md5,
                    'config': config
                }
                json.dump(meta, f)

        except KeyboardInterrupt:
            controller['running'] = False
            sys.exit(1)
    else:
        print "Downloaded images up2date in " + dir
        print " - Remove this directory if you want to re-download all images of your dataset and re-shuffle training/validation images."

    trainer.output_size = len(classes)
    trainer.set_status('LOAD IMAGE DONE')

    # change to type local_images
    dataset_transformed = dataset.copy()
    dataset_transformed['config']['path'] = dir

    all_memory = get_option(dataset['config'], 'allMemory', False, 'bool')

    if all_memory:
        return read_images_in_memory(job_config, dataset_transformed, node, trainer)
    else:
        return read_images_keras_generator(job_config, dataset_transformed, node, trainer)


from __future__ import division

import base64
import os
import time
import cStringIO

import PIL.Image
import math
from keras.callbacks import Callback
from keras import backend as K
import keras.layers.convolutional

import numpy as np

from aetros.utils.image import get_layer_vis_square
from network import ensure_dir

class KerasLogger(Callback):
    def __init__(self, trainer, backend, job_model, general_logger):
        super(KerasLogger, self).__init__()
        self.validation_per_batch = []
        self.ins = None

        self.trainer = None

        self.trainer = trainer
        self.backend = backend
        self.job_model = job_model
        self.general_logger = general_logger
        self._test_with_acc = None
        self.last_batch_time = time.time()
        self.start_time = time.time()
        self.made_batches = 0
        self.batches_per_second = 0
        self.stats = []
        self.last_current = None
        self.current = {}
        self.filepath_best = job_model.get_weights_filepath_best()
        self.filepath_latest = job_model.get_weights_filepath_latest()

        ensure_dir(os.path.dirname(self.filepath_best))

        self.insight_sample_input_item = None

        self.best_epoch = 0
        self.best_total_accuracy = 0
        self.worst_total_accuracy = 0

    def set_trainer(self, trainer):
        self.trainer = trainer
        self.current['epoch'] = 0
        self.current['started'] = self.start_time
        self.trainer.set_job_info('current', self.current)

    def on_train_begin(self, logs={}):
        self.start_time = time.time()
        self.last_batch_time = time.time()

        # pprint('on_train_begin')
        # pprint(logs)
        # pprint(self.params)

        nb_sample = self.params['nb_sample'] #training samples total
        nb_epoch = self.params['nb_epoch'] #training samples total

        self.current['nb_sample'] = nb_sample
        self.current['nb_epoch'] = nb_epoch

    def on_batch_begin(self, batch, logs={}):

        # pprint('on_batch_begin')
        # pprint(logs)
        # pprint(self.current)

        batch_size = logs['size']
        nb_batches = math.ceil(self.current['nb_sample'] / batch_size) #normal nb batches

        self.current['nb_batches'] = nb_batches
        self.current['batch_size'] = batch_size

    def on_batch_end(self, batch, logs={}):
        loss = logs['loss']
        if math.isnan(loss) or math.isinf(loss):
            loss = -1

        # pprint('on_batch_end %d ' % (batch,))
        # pprint(logs)
        # pprint(self.current)

        self.validation_per_batch.append(loss)

        current_batch = logs['batch']
        current_batch_size = logs['size'] #how many training items in this batch, differs for the last run

        self.made_batches += 1

        time_diff = time.time() - self.last_batch_time

        if time_diff > 1 or batch == self.current['nb_batches']: #only each second or last batch
            self.batches_per_second = self.made_batches / time_diff
            self.made_batches = 0
            self.last_batch_time = time.time()

            nb_sample = self.params['nb_sample'] #training samples total
            nb_epoch = self.params['nb_epoch'] #training samples total
            batch_size = self.current['batch_size'] #normal batch size
            nb_batches = nb_sample / batch_size #normal nb batches

            self.current['batchesPerSecond'] = self.batches_per_second
            self.current['itemsPerSecond'] = self.batches_per_second * current_batch_size

            epochs_per_second = self.batches_per_second / nb_batches #all batches
            self.current['epochsPerSecond'] = epochs_per_second

            made_batches = (self.current['epoch'] * nb_batches) + current_batch
            total_batches = nb_epoch * nb_batches
            needed_batches = total_batches - made_batches
            seconds_per_batch = 1 / self.batches_per_second

            self.current['eta'] = seconds_per_batch * needed_batches

            self.current['currentBatch'] = current_batch
            self.current['currentBatchSize'] = current_batch_size
            elapsed = time.time() - self.start_time
            self.current['elapsed'] = elapsed

            self.trainer.set_job_info('current', self.current)

    def write(self, line):
        self.general_logger.write(line)

    def on_epoch_end(self, epoch, logs={}):
        log = logs.copy()

        self.filterInvalidJsonValues(log)

        log['created'] = time.time()
        log['epoch'] = epoch+1
        log['loss'] = sum(self.validation_per_batch) / float(len(self.validation_per_batch))

        self.validation_per_batch = []
        log['validation_accuracy'] = {}
        log['validation_loss'] = {}
        log['training_loss'] = {}
        log['training_accuracy'] = {}

        trainer = self.trainer
        elapsed = time.time() - self.start_time

        total_loss = 0
        total_accuracy = 0

        for layer in trainer.model.outputs:
            #todo, this is not very generic
            log['validation_loss'][layer.name] = log['val_loss'] #outs[0]
            log['validation_accuracy'][layer.name] = log['val_acc'] #outs[1]

            log['training_loss'][layer.name] = log['loss'] #outs[0]
            log['training_accuracy'][layer.name] = log['acc'] #outs[1]

            total_loss += log['val_loss']
            total_accuracy += log['val_acc']

        if total_accuracy > self.best_total_accuracy:
            self.best_total_accuracy = total_accuracy
            self.best_epoch = log['epoch']
            self.model.save_weights(self.filepath_best, overwrite=True)

        self.model.save_weights(self.filepath_latest, overwrite=True)

        if total_accuracy < self.worst_total_accuracy:
            self.worst_total_accuracy = total_accuracy

        self.current['totalValidationLoss'] = total_loss
        self.current['totalValidationAccuracy'] = total_accuracy
        self.current['totalValidationAccuracyBest'] = self.best_total_accuracy
        self.current['totalValidationAccuracyWorst'] = self.worst_total_accuracy
        self.current['totalValidationAccuracyBestEpoch'] = self.best_epoch

        self.current['totalTrainingLoss'] = log['loss']
        self.current['elapsed'] = elapsed
        self.current['epoch'] = log['epoch']

        self.filterInvalidJsonValues(self.current)

        trainer.set_job_info('current', self.current)

        line = "Epoch %d: loss=%f, acc=%f, val_loss=%f, val_acc=%f\n" % (log['epoch'], log['loss'], log.get('acc'), log['val_loss'], log.get('val_acc'), )
        self.general_logger.write(line)

        self.backend.job_add_status('epoch', log)

        input_name = None

        if self.job_model.job['insights']:

            #Todo, support multiple inputs
            first_input = self.model.inputs[0]
            first_input_layer = self.model.input_layers[0]
            first_output_layer = self.model.output_layers[0]

            if first_input_layer != None:
                input_data_x = self.trainer.data_validation['x'][first_input_layer.name]
                input_data_y = self.trainer.data_validation['y'][first_output_layer.name]
                confusion_matrix = {}

                if self.insight_sample_input_item is None:
                    if self.trainer.is_generator(input_data_x):
                        batch_x, batch_y = input_data_x.next()
                        self.insight_sample_input_item = batch_x[0]
                    else:
                        self.insight_sample_input_item = input_data_x[0]

                # build confusion matrix
                node = self.job_model.get_model_node(first_output_layer.name)
                if node['classificationMode'] == 'categorical':
                    matrix = np.zeros((len(trainer.classes), len(trainer.classes)))
                    if self.trainer.is_generator(input_data_x):
                        processed_samples = 0

                        while processed_samples < trainer.nb_val_samples:
                            generator_output = input_data_x.next()
                            if len(generator_output) == 2:
                                x, y = generator_output
                                sample_weight = None
                            elif len(generator_output) == 3:
                                x, y, sample_weight = generator_output
                            else:
                                self.model._stop.set()
                                raise Exception('output of generator should be a tuple '
                                                '(x, y, sample_weight) '
                                                'or (x, y). Found: ' + str(generator_output))

                            if type(x) is list:
                                nb_samples = len(x[0])
                            elif type(x) is dict:
                                nb_samples = len(list(x.values())[0])
                            else:
                                nb_samples = len(x)

                            processed_samples += nb_samples

                            prediction = trainer.model.predict_on_batch(x)
                            predicted_classes = prediction.argmax(axis=-1)
                            expected_classes = y.argmax(axis=-1)

                            for sample_idx, predicted_class in enumerate(predicted_classes):
                                expected_class = expected_classes[sample_idx]
                                matrix[expected_class, predicted_class] += 1

                    else:
                        prediction = trainer.model.predict(input_data_x, batch_size=self.job_model.get_batch_size())
                        predicted_classes = prediction.argmax(axis=-1)
                        expected_classes = input_data_y.argmax(axis=-1)

                        for sample_idx, predicted_class in enumerate(predicted_classes):
                            expected_class = expected_classes[sample_idx]
                            matrix[expected_class, predicted_class] += 1

                    confusion_matrix[first_output_layer.name] = matrix.tolist()

                images = []

                try:
                    image = self.make_image(self.insight_sample_input_item)
                    images.append({
                        'id': input_name,
                        'title': input_name,
                        'image': self.to_base64(image)
                    })
                except:
                    pass

                uses_learning_phase = self.model.uses_learning_phase
                inputs = [first_input]
                input_data_x_sample = [[self.insight_sample_input_item]]

                if uses_learning_phase:
                    inputs += [K.learning_phase()]
                    input_data_x_sample += [0.]  # disable learning_phase

                for layer in self.model.layers:
                    if isinstance(layer, keras.layers.convolutional.Convolution2D) or isinstance(layer, keras.layers.convolutional.MaxPooling2D):

                        fn = K.function(inputs, [layer.output])
                        Y = fn(input_data_x_sample)[0]

                        data = np.squeeze(Y)
                        # print("Layer Activations " + layer.name)
                        image = PIL.Image.fromarray(get_layer_vis_square(data))

                        images.append({
                            'id': layer.name,
                            'type': 'convolution',
                            'title': layer.name,
                            'image': self.to_base64(image)
                        })

                        if hasattr(layer, 'W') and layer.W:
                            # print("Layer Weights " + layer.name)
                            data = layer.W.get_value()
                            image = PIL.Image.fromarray(get_layer_vis_square(data))
                            images.append({
                                'id': layer.name+'_weights',
                                'type': 'convolution',
                                'title': layer.name + ' weights',
                                'image': self.to_base64(image)
                            })


                    if isinstance(layer, keras.layers.Dense):

                        fn = K.function(inputs, [layer.output])
                        Y = fn(input_data_x_sample)[0]

                        node = self.job_model.get_model_node(layer.name)
                        if node and node['activationFunction'] == 'softmax':
                            image = self.make_image_from_dense_softmax(np.squeeze(Y))
                        else:
                            image = self.make_image_from_dense(np.squeeze(Y))

                        images.append({
                            'id': layer.name,
                            'type': 'dense',
                            'title': layer.name,
                            'image': self.to_base64(image)
                        })

                self.backend.job_add_insight({'epoch': log['epoch'], 'confusionMatrix': confusion_matrix}, images)



    def filterInvalidJsonValues(self, dict):
        for k,v in dict.iteritems():
            if math.isnan(v) or math.isinf(v):
                dict[k] = -1

    def to_base64(self, image):
        buffer = cStringIO.StringIO()
        image.save(buffer, format="JPEG", optimize=True, quality=80)
        return base64.b64encode(buffer.getvalue())

    def make_image(self, data):
        from keras.preprocessing.image import array_to_img
        try:
            image = array_to_img(data)
        except:
            return None

        # image = image.resize((128, 128))

        return image

    def make_image_from_dense_softmax(self, neurons):
        from aetros.utils import array_to_img

        img = array_to_img(neurons.reshape((1, len(neurons), 1)))
        img = img.resize((9, len(neurons)*8))

        return img

    def make_image_from_dense(self, neurons):
        from aetros.utils import array_to_img
        cols = int(math.ceil(math.sqrt(len(neurons))))

        even_length = cols*cols
        diff = even_length - len(neurons)
        if diff > 0:
            neurons = np.append(neurons, np.zeros(diff, dtype=neurons.dtype))

        img = array_to_img(neurons.reshape((1, cols, cols)))
        img = img.resize((cols*8, cols*8))

        return img



import sys
import time
import datetime

class GeneralLogger():
    def __init__(self, job, logFD, aetros_backend, error=False):
        self.error = error
        self.job = job
        self.aetros_backend = aetros_backend

        self.terminal = sys.stdout if error is False else sys.stderr
        self.logFD = logFD

    def get_time(self):
        ts = time.time()
        st = datetime.datetime.fromtimestamp(ts).strftime('%Y-%m-%d %H:%M:%S')

        return st + '.' + str(ts%1)[2:6]

    def get_line(self, line):
        if line:
            return "[%s] %s\n" % (self.get_time(), line)

        return line

    def flush(self):
        self.logFD.flush()
        self.terminal.flush()

    def write(self, message):

        # if message == '\n':
        #     return

        message = unicode(message)

        # if not self.error:
        #     message = self.get_line(message)

        self.logFD.write(message)
        self.terminal.write(message)
        self.aetros_backend.write_log(self.job['id'], message)



from __future__ import division
import time
import psutil
from threading import Thread
import numpy as np

class MonitoringThread(Thread):
    def __init__(self, job, aetros_backend, trainer):
        Thread.__init__(self)
        self.trainer = trainer
        self.job = job
        self.aetros_backend = aetros_backend
        self.job_id = job['id']
        self.second = 0
        self.running = True

    def stop(self):
        self.running = False

    def run(self):
        while self.running:
            self.monitor()
            time.sleep(0.1)

    def monitor(self):
        cpu_util = np.mean(psutil.cpu_percent(interval=1, percpu=True))
        mem = psutil.virtual_memory()

        gpu_memory_use = None

        if self.trainer.on_gpu:
            from theano.sandbox import cuda
            if cuda.cuda_ndarray.cuda_ndarray.mem_info:
                gpu = cuda.cuda_ndarray.cuda_ndarray.mem_info()
                gpu_memory_use = (gpu[1] - gpu[0]) / gpu[1] * 100

        self.aetros_backend.job_add_status('system', {
            'second': self.second,
            'cpu': cpu_util,
            'memory': mem.percent,
            'memory_gpu': gpu_memory_use
        })

        self.second += 1
        pass


import argparse
import os
import sys

import aetros.const

class StartCommand:

    def main(self, args):

        from aetros.starter import start
        parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter, prog=aetros.const.__prog__ + ' start')
        parser.add_argument('network_name', nargs='?', help='the network name, e.g. aetros/mnist-network')
        parser.add_argument('--insights', action='store_true', help="activates insights")
        parser.add_argument('--dataset', help="Dataset id when network has placeholders")
        parser.add_argument('--gpu', action='store_true', help="Activates GPU if available")
        parser.add_argument('--device', help="Which device index should be used. Default 0 (which means with --gpu => 'gpu0')")
        parser.add_argument('--tf', action='store_true', help="Uses TensorFlow instead of Theano")
        parser.add_argument('--mp', help="Activates multithreading if available with given thread count.")

        parsed_args = parser.parse_args(args)

        if not parsed_args.network_name:
            parser.print_help()
            sys.exit()

        import os
        flags = os.environ['THEANO_FLAGS'] if 'THEANO_FLAGS' in os.environ else ''
        if parsed_args.gpu:
            if parsed_args.device:
                flags += ",device=gpu" + parsed_args.device
            else:
                flags += ",device=gpu"

        if parsed_args.mp:
            flags += ",openmp=True"
            os.environ['OMP_NUM_THREADS'] = parsed_args.mp

        os.environ['THEANO_FLAGS'] = flags

        if parsed_args.tf:
            os.environ['KERAS_BACKEND'] = 'tensorflow'

        start(parsed_args.network_name, dataset_id=parsed_args.dataset, insights=parsed_args.insights)



from __future__ import print_function, division

import io
import logging
import os
import pprint
import signal
import sys
import traceback

import network
from AetrosBackend import AetrosBackend
from GeneralLogger import GeneralLogger
from JobModel import JobModel
from MonitorThread import MonitoringThread
from Trainer import Trainer
from network import ensure_dir


def start(job_id, dataset_id=None, server_id='local', insights=False):
    """
    Starts the training process with all logging of a job_id
    """

    aetros_backend = AetrosBackend(job_id)

    if '/' in job_id:
        print("...")
        job_id = aetros_backend.create_job(job_id, server_id=server_id, dataset_id=dataset_id, insights=insights)
        if job_id is None:
            exit(1)

        print("Training '%s' created and started. Open http://%s/trainer/app?training=%s to monitor the training." %
              (job_id, aetros_backend.host, job_id))
    else:
        print("Training '%s' restarted. Open http://%s/trainer/app?training=%s to monitor the training." %
              (job_id, aetros_backend.host, job_id))

    aetros_backend.job_id = job_id
    job = aetros_backend.get_job()
    if job is None:
        exit(1)

    if job == 'Job not found':
        raise Exception('Training not found. Have you configured your token correctly?')

    if not isinstance(job, dict):
        raise Exception('Training does not exist. Make sure you created the job via AETROS TRAINER')

    if not len(job['config']):
        raise Exception('Training does not have a configuration. Make sure you created the job via AETROS TRAINER')

    network_id = job['networkId']

    aetros_backend.job_started(job_id, os.getpid())

    ensure_dir('networks/%s/%s' % (network_id, job_id))

    log = io.open('networks/%s/%s/network.log' % (network_id, job_id), 'w', encoding='utf8')
    log.truncate()

    job_model = JobModel(aetros_backend, job)
    general_logger = GeneralLogger(job, log, aetros_backend)

    print("start network ...")

    from KerasLogger import KerasLogger
    trainer = Trainer(aetros_backend, job_model, general_logger)
    keras_logger = KerasLogger(trainer, aetros_backend, job_model, general_logger)
    trainer.callbacks.append(keras_logger)

    sys.stdout = general_logger
    sys.stderr = general_logger

    job['running'] = True

    monitoringThread = MonitoringThread(job, aetros_backend, trainer)
    monitoringThread.daemon = True
    monitoringThread.start()
    network.collect_system_information(trainer)

    def ctrlc(sig, frame):
        print("signal %s received\n" % id)
        raise KeyboardInterrupt("CTRL-C!")

    signal.signal(signal.SIGINT, ctrlc)

    try:
        print("Setup training")
        network.job_prepare(job)

        print("Start training")
        network.job_start(job_model, trainer, keras_logger, general_logger)

        job['running'] = False
        job_model.sync_weights()
        aetros_backend.stop_syncer()
        aetros_backend.post('job/stopped', json={'id': job_model.id, 'status': 'DONE'})

        print("done.")
        sys.exit(0)
    except KeyboardInterrupt:
        trainer.set_status('STOPPING')
        print('Early stopping ...')

        if aetros_backend.stop_requested:
            print(' ... stop requested through trainer.')

        if trainer.model:
            trainer.model.stop_training = True

        monitoringThread.stop()
        job_model.sync_weights()
        aetros_backend.stop_syncer()
        aetros_backend.post('job/stopped', json={'id': job_model.id, 'status': 'EARLY STOP'})
        print("out.")
        sys.exit(1)
    except Exception as e:
        print("Crashed ...")

        if trainer.model:
            trainer.model.stop_training = True

        log.write(unicode(traceback.format_exc()))
        logging.error(traceback.format_exc())

        monitoringThread.stop()
        aetros_backend.stop_syncer()
        aetros_backend.post('job/stopped', json={'id': job_model.id, 'status': 'CRASHED', 'error': e.message})
        print("out.")
        raise e



import os
import sys

class JobModel:
    def __init__(self, backend, job):
        self.backend = backend
        self.job = job

        network_id = job['networkId']
        job_id = job['id']
        self.id = job['id']

        self.network_id = network_id
        self.job_id = job_id
        pass

    def get_model_node(self, name):
        for nodes in self.job['config']['layer']:
            for node in nodes:
                if node['name'] == name:
                    return node

        raise Exception('Could not found node name=%s in model.' % (name,))

    def get_batch_size(self):
        return self.job['config']['settings']['batchSize']

    def get_network_h5_path(self):
        return os.getcwd() + '/networks/%s/%s/network.h5' % (self.network_id, self.job_id)

    def get_dataset_dir(self):
        return os.getcwd() + '/networks/%s/%s/datasets' % (self.network_id, self.job_id)

    def get_base_dir(self):
        return os.getcwd() + '/networks/%s/%s' % (self.network_id, self.job_id)

    def get_dataset_downloads_dir(self, dataset):
        return os.getcwd() + '/datasets/%s/datasets_downloads' % (dataset['id'],)

    def get_weights_filepath_latest(self):
        return os.getcwd() + '/weights/%s/latest.hdf5' % (self.job_id,)

    def get_weights_filepath_best(self):
        return os.getcwd() + '/weights/%s/best.hdf5' % (self.job_id,)

    def get_model_provider(self):

        sys.path.append(self.get_base_dir())

        import model_provider
        print "Imported model_provider in %s " % (self.get_base_dir() + '/model_provider.py',)
        sys.path.pop()

        return model_provider

    def sync_weights(self):
        self.backend.job_add_status('status', 'SYNC WEIGHTS')
        print "Sync weights ..."
        self.backend.upload_weights('latest', self.get_weights_filepath_latest())
        self.backend.upload_weights('best', self.get_weights_filepath_best())
        print "Weights synced."

    def network_get_datasets(self, trainer):
        datasets_dir = self.get_dataset_dir()

        datasets = {}

        from aetros.utils import get_option
        from auto_dataset import get_images, read_images_keras_generator, read_images_in_memory

        # load placeholder, auto data
        config = self.job['config']
        for net in config['layer'][0]:
            if 'datasetId' in net and net['datasetId']:

                dataset = config['datasets'][net['datasetId']]
                if not dataset:
                    raise Exception('Dataset of id %s does not exists. Available %s' % (net['datasetId'], ','.join(config['datasets'].keys())))

                if dataset['type'] == 'images_upload' or dataset['type'] == 'images_search':

                    connected_to_net = self.get_connected_network(config['layer'], net)
                    if connected_to_net == None:
                        # this input is not in use, so we dont need to calculate its dataset
                        continue

                    datasets[net['datasetId']] = get_images(config, dataset, net, trainer)

                elif dataset['type'] == 'images_local':

                    all_memory = get_option(dataset['config'], 'allMemory', False, 'bool')

                    if all_memory:
                        datasets[net['datasetId']] = read_images_in_memory(config, dataset, net, trainer)
                    else:
                        datasets[net['datasetId']] = read_images_keras_generator(config, dataset, net, trainer)

                elif dataset['type'] == 'python':
                    name = dataset['id'].replace('/', '__')

                    sys.path.append(datasets_dir)
                    data_provider = __import__(name, '')
                    print "Imported dataset provider in %s " % (datasets_dir + '/' + name + '.py', )
                    sys.path.pop()
                    datasets[dataset['id']] = data_provider.get_data()


        return datasets

    def get_first_input_layer(self):
        config = self.job['config']
        return config['layer'][0][0]

    def get_first_output_layer(self):
        config = self.job['config']
        return config['layer'][-1][0]

    def get_dataset(self, dataset_id):
        return self.job['config']['datasets'][dataset_id]

    def get_connected_network(self, layers, to_net):
        connected_to_net = None
        for nets in layers:
            for net in nets:
                if 'connectedTo' in net:
                    for connectedTo in net['connectedTo']:
                        if connectedTo == to_net['id']:
                            return net

        return connected_to_net


from __future__ import print_function

import io
import json
import tempfile
import urllib

import numpy as np
import sys
import os

from PIL import Image

from aetros import network
from aetros.network import ensure_dir
from JobModel import JobModel
from AetrosBackend import AetrosBackend


def predict(job_id, file_path, insights=False):
    print("Prepare network ...")
    aetros_backend = AetrosBackend(job_id)

    job = aetros_backend.get_light_job()
    job_id = job['id']

    if job == 'Job not found':
        raise Exception('Job not found. Have you configured your token correctly?')

    if not isinstance(job, dict):
        raise Exception('Job does not exist. Make sure you created the job via AETROS TRAINER')

    if not len(job['config']):
        raise Exception('Job does not have a configuration. Make sure you created the job via AETROS TRAINER')

    network_id = job['networkId']
    log = io.open(tempfile.mktemp(), 'w', encoding='utf8')
    log.truncate()

    ensure_dir('networks/%s/%s' % (network_id, job_id))
    network.job_prepare(job)

    job_model = JobModel(aetros_backend, job)

    weight_path = job_model.get_weights_filepath_best()
    if not os.path.exists(weight_path) or os.path.getsize(weight_path) == 0:
        weight_url = aetros_backend.get_best_weight_url(job_id)
        if not weight_url:
            print("No weights available for this job.")
            exit(1)

        print("Download weights %s to %s .." % (weight_url, weight_path))
        ensure_dir(os.path.dirname(weight_path))

        f = open(weight_path, 'wb')
        f.write(urllib.urlopen(weight_url).read())
        f.close()

    from GeneralLogger import GeneralLogger
    from Trainer import Trainer

    general_logger = GeneralLogger(job, log, aetros_backend)
    trainer = Trainer(aetros_backend, job_model, general_logger)

    first_input_layer = job_model.get_first_input_layer()

    size = (first_input_layer['width'], first_input_layer['height'])
    if first_input_layer['inputType'] == 'image':
        trainer.input_shape = (1, size[0], size[1])
        grayscale = True
    elif first_input_layer['inputType'] == 'image_rgb':
        grayscale = False
        trainer.input_shape = (3, size[0], size[1])
    else:
        trainer.input_shape = (size[0] * size[1],)
        grayscale = True

    first_output_layer = job_model.get_first_output_layer()
    output_dataset = job_model.get_dataset(first_output_layer['datasetId'])

    if output_dataset:
        if 'classes' in job['info']:
            trainer.output_size = len(job['info']['classes'])

    model_provider = job_model.get_model_provider()
    model = model_provider.get_model(trainer)

    loss = model_provider.get_loss(trainer)
    optimizer = model_provider.get_optimizer(trainer)
    model_provider.compile(trainer, model, loss, optimizer)
    model.load_weights(weight_path)

    if 'http://' in file_path or 'https://' in file_path:
        local_image_path = tempfile.mktemp()
        print("Download input ...")
        f = open(local_image_path, 'wb')
        f.write(urllib.urlopen(file_path).read())
        f.close()
    else:
        local_image_path = file_path

    image = Image.open(local_image_path)
    image = image.resize(size, Image.ANTIALIAS)

    if grayscale:
        image = image.convert("L")
    else:
        image = image.convert("RGB")

    image = np.asarray(image, dtype='float32')

    if len(trainer.input_shape) > 1:
        # RGB: height, width, channel -> channel, height, width
        image = image.transpose(2, 0, 1)
    else:
        # L: height, width => height*width
        image = image.reshape(size[0] * size[1])

    image = image / 255

    input = {}
    for input_layer in model.input_layers:
        input[input_layer.name] = np.array([image])

    print("Start prediction ...")
    prediction = model.predict(input)

    output = dict(zip(job['info']['classes'], prediction[0].tolist()))
    output = sorted(output.items(), reverse=True, key=lambda (k, v): v)
    print(json.dumps(output, indent=4))


import argparse
import sys

import aetros.const

class PredictCommand:

    def main(self, args):

        from aetros.predict import predict
        parser = argparse.ArgumentParser(formatter_class=argparse.RawTextHelpFormatter, prog=aetros.const.__prog__ + ' predict')
        parser.add_argument('job', nargs='?', help='the job id, e.g. GoVDO1Njm')
        parser.add_argument('--insights', action='store_true', help="activates insights")
        parser.add_argument('-i', nargs='+', help="Input (path or url). Multiple allowed")

        parsed_args = parser.parse_args(args)

        if not parsed_args.job:
            parser.print_help()
            sys.exit()

        predict(parsed_args.job, parsed_args.i[0], parsed_args.insights)



from setuptools import setup
from setuptools import find_packages
import aetros

setup(name='aetros',
      version=aetros.const.__version__,
      description='Console application to manage deep neural network trainings in AETROS TRAINER',
      author='Marc J. Schmidt',
      author_email='marc@marcjschmidt.de',
      url='https://github.com/aetros/aetros-cli',
      download_url='https://github.com/aetros/worker/tarball/0.1.0',
      license='MIT',
          entry_points = {
              'console_scripts': ['aetros = aetros:main'],
          },
      install_requires=(
            'keras>=1.0.6',
            'requests',
            'numpy',
            'scipy',
            'h5py',
            'py-cpuinfo==0.2.3'
      ),
      packages=find_packages())